"""Aggregation of the precise parser and the byline parser,
leading to high-level and complete parsing.

"""

from clyngor import asp_parsing
from clyngor.asp_parsing import precise_parser
from clyngor.asp_parsing import byline_parser



